#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/times.h>
#include <sys/time.h>

typedef struct TimeInfo_s {
    struct tms initial_tms;
    struct tms stamp_tms;
    time_t timeStamp;
    time_t startTime;
} TimeInfo_t;

#define DAY 86400	/* seconds in a day */
#define HOUR 3600	/* seconds in an hour */
#define MINUTE 60	/* seconds in a minute */

#define BLK_SIZE 2048

TimeInfo_t timeInfo;

FILE *f;

int updateMode = 1;
int blk_header[16];
int blk[BLK_SIZE];

/* use 64-bit file sizes/offsets */
off64_t curPos = 0;
off64_t lastPos = 0;
off64_t blk_start = -1;
off64_t ref_start = -1;
off64_t ref_end = -1;

char buf[BUFSIZ];

/* Start keeping time. */
void StartTime(void) {
    times(&timeInfo.initial_tms);
    timeInfo.timeStamp = timeInfo.startTime = time(0);
}

/* Calculate time data. */
void calcTimeInfo(time_t start_time, int *end_time,
                  struct tms *start_tms, struct tms *end_tms,
                  char *in_str, char *out_str, int delta) {
    int eDays, eHour, eMin, eSec;
    int eTime;
    char tmp[1024];
    double tick_sec = sysconf(_SC_CLK_TCK);
    double uSec, sSec;

    /* Calculate relative or absolute elapsed time. */
    if (!delta) {
        *end_time = time(0);
    }
    eTime = *end_time - start_time;
    eDays = eTime / DAY;    eTime %= DAY;
    eHour = eTime / HOUR;   eTime %= HOUR;
    eMin  = eTime / MINUTE;
    eSec  = eTime % MINUTE;

    /* Generate relative or absolute time string. */
    if (!delta) {
        out_str[0] = '\0';
    }
    strcat(out_str, in_str);
    if (eDays) {
        sprintf(tmp,"%d ", eDays);
        strcat(out_str, tmp);
    }
    sprintf(tmp, "%02d:%02d:%02d", eHour, eMin, eSec);
    strcat(out_str, tmp);

    /* Calculate user and sys times (relative or absolute). */
    if (!delta) {
        times(end_tms);
    }
    uSec = end_tms->tms_utime - start_tms->tms_utime;
    sSec = end_tms->tms_stime - start_tms->tms_stime;
    uSec /= tick_sec;
    sSec /= tick_sec;

    /* Append user+sys time string. */
    sprintf(tmp,"  User=%.2f Sys=%.2f", uSec, sSec);
    strcat(out_str, tmp);
}
    
/* Report absolute and delta usr/sys times. */
char *ReportTime(char *input_string) {
    int cTime;
    struct tms end_tms;
    static char time_str[1024];

    calcTimeInfo(timeInfo.startTime, &cTime, &timeInfo.initial_tms, &end_tms, input_string, time_str, 0);
    calcTimeInfo(timeInfo.timeStamp, &cTime, &timeInfo.stamp_tms,   &end_tms, " (delta: ",  time_str, 1);

    timeInfo.stamp_tms = end_tms;
    timeInfo.timeStamp = cTime;

    return(time_str);
}

/* Open the next block.
 * Output the header, return the start-of-block offset.
 */
off64_t open_blk(FILE *f)
{
    /* Get current offset into the file. */
    off64_t cur = ftello64(f);

    /* Write this block's header. */
    fwrite(blk_header, sizeof(blk_header), 1, f);

    return (cur);
}

/* If in update mode, seek back and update the block header. */
void update_blk(FILE *f, off64_t loc) {
    /* Punt if we're not in update mode. */
    if (!updateMode)
        return;

    /* Determine where we need to seek back to after the update. */
    if (ref_start < 0) {
        ref_start = ftello64(f);
        ref_end = ref_start;
    } else {
        ref_end = ftello64(f);
    }

    /* Seek back to block start-of-block. */
    fseeko64(f, loc, SEEK_SET);

    /* Overwrite with updated information. */
    fwrite(blk_header, sizeof(blk_header), 1, f);

    /* Seek back to present position. */
    fseeko64(f, ref_end, SEEK_SET);
}

/* Close the block, updating if necessary. */
void close_blk(FILE *f, off64_t loc)
{
    update_blk(f, loc);
}

/* Write the current block to disk. */
void write_blk(FILE *f)
{
    int size, i;

    /* blk size is random, between 1-BLK_SIZE */
    size = (rand() % BLK_SIZE) + 1;

    /* Output how many. */
    fwrite(&size, sizeof(int), 1, f);

    /* Output address=data. */
    for(i = 0; i < size; i++) {
        fwrite(&i, sizeof(int), 1, f);
    }
}

int main(int argc, char* argv[])
{
    int i, j, num_blks;
    char *filename;

    if (argc != 4) {
        printf("Usage: %s num_blks filename mode [0: non-update 1: update]\n", argv[0]);
        return 1;
    }

    num_blks   = atoi(argv[1]);
    filename   = argv[2];
    updateMode = atoi(argv[3]);

    printf("blks=%d file=%s bufsize=%d\n", num_blks, filename, sizeof(buf));

    /* Start keeping user/sys times. */
    StartTime();

    /* Begin writing to output file. */
    f = fopen64(filename, "wb");        /* use 64-bit file sizes */
    setbuf(f, buf);

    for (i = 0; i < num_blks; i++) {
        /* Initialized block header data. */
        for(j = 0; j < 16; j++)
            blk_header[j] = (2 * j);

        /* Open block for writing, get starting position. */
        blk_start = open_blk(f); 

        /* Write block.  Duh.  */
        write_blk(f);

        /* Update the header contents. */
        for(j = 0; j < 16; j++)
            blk_header[j] = j;

        /* Close the block (update if in updateMode). */
        close_blk(f, blk_start);

        /* Output timestamps every 32k blocks (absolute and relative). */
        if (!(i % 32768)) {
            curPos = ftello64(f);
            printf("  Row: %7d  [%s File=%d)]\n", i, ReportTime(" "), (curPos - lastPos));
            lastPos = curPos;
        }
    }

    fclose(f);
    return 0;
}
